#!/usr/bin/env python3

from sys import platform
from argparse import ArgumentParser
from contextlib import redirect_stdout
from os import devnull
from sys import stdout, stderr, exit
from pathlib import Path
from woven import WovenMesh
from json import JSONDecodeError
from cattrs.errors import ClassValidationError
from tkinter.filedialog import askopenfilename
from tkinter.messagebox import showerror
import customtkinter as ctk

def load_config(config_path: str) -> WovenMesh | None:
    try:
        return WovenMesh.load_json_file(config_path)
    except FileNotFoundError:
        pass
    except JSONDecodeError as e:
        showerror("Invalid JSON Error", f"Invalid JSON encountered in configuration file: {e}")
    except ClassValidationError as e:
        details = "\n".join(f'{type(e).__name__}: {e}' for e in e.exceptions)
        showerror("Validation Error", f"The following validation errors occurred when loading the configuration file:\n{details}")
    return None

def start(config_path: str):
    ctk.set_appearance_mode("System")
    ctk.set_default_color_theme("dark-blue")

    if platform != "darwin" and not platform.startswith("win"):
        ctk.set_widget_scaling(1.25)
        ctk.set_window_scaling(1.25)

    root = ctk.CTk()
    root.geometry("1200x800")
    root.title("Woven")
    root.minsize(240, 240)
    root.grid_columnconfigure(0, weight = 1)
    root.grid_rowconfigure(1, weight = 1)

    def update_config_handler():
        fn = askopenfilename(filetypes = [("JSON Files", "*.json")])
        if fn:
            update_config(load_config(fn))
    
    load_config_button = ctk.CTkButton(master = root, text = "Load Config", width = 0, command = update_config_handler)
    load_config_button.grid(row = 0, padx = 20, pady = (20, 10), sticky = "w")
    save_config_button = ctk.CTkButton(master = root, text = "Save Config", width = 0, command = update_config_handler)
    save_config_button.grid(row = 0, padx = 20, pady = (20, 10), sticky = "e")
    
    config_textbox = ctk.CTkTextbox(master = root)
    config_textbox.grid(row = 1, column = 0, padx = 20, pady = (10, 20), sticky = "nsew")

    def update_config(config: WovenMesh | None):
        config_textbox.delete("0.0", "end")
        config_textbox.insert("0.0", "" if config is None else config.to_json_str())
    
    update_config(load_config(config_path))
    
    root.mainloop()

def main():
    parser = ArgumentParser("woven-ui")
    parser.add_argument("-q", "--quiet", action = "store_true", help = "decrease output verbosity")
    parser.add_argument("-c", "--config", default = "config.json", help = "The path to the config file")
    args = parser.parse_args()
    
    with redirect_stdout(open(devnull, "w") if args.quiet else stdout):
        start(args.config)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        exit(130)
    